from __future__ import annotations

from rdflib import URIRef


class Entity:
    """
    Abstraction of a KG entity with basic RDF properties plus its parent_entity (connected in KG with rdf:type).
    """

    def __init__(self, iri: str, parent_entity: Entity = None):
        self.iri = URIRef(iri)
        self.parent_entity = parent_entity
        self.namespace = self.get_namespace(iri)
        self.name = self.type = self.get_descriptor(iri)
        if parent_entity:
            self.type = parent_entity.name

    @staticmethod
    def get_namespace(iri: str) -> str:
        return iri.split("#")[0]

    @staticmethod
    def get_descriptor(iri: str) -> str:
        return iri.split("#")[1]



from typing import Callable, List, Optional, Tuple

from classes.entity import Entity
from rdflib import Graph, Namespace, URIRef, query


def query_method_parent_classes(kg, method_iri):
    return kg.query(
        f"SELECT ?c WHERE {{ ?method rdfs:subClassOf ?c . }}",
        initBindings={"method": URIRef(method_iri)},
    )


def query_entity_parent_iri(kg, entity_iri: str, upper_class_uri_ref: URIRef):
    return kg.query(
        f"SELECT ?t WHERE {{ ?entity rdf:type ?t ." f"                   ?t rdfs:subClassOf* ?upper_class .}}",
        initBindings={
            "entity": URIRef(entity_iri),
            "upper_class": upper_class_uri_ref,
        },
    )


def query_data_entity_reference_iri(kg, namespace_prefix, entity_iri: str):
    return kg.query(
        f"SELECT ?r WHERE {{ ?entity {namespace_prefix}:hasReference ?r . }}",
        initBindings={
            "entity": URIRef(entity_iri),
        },
    )


def query_method_iri_by_task_iri(kg, namespace_prefix, task_iri: str):
    return kg.query(
        f"SELECT ?m WHERE {{ ?task ?m_property ?m ."
        f"                   ?m_property rdfs:subPropertyOf* {namespace_prefix}:hasMethod .}}",
        initBindings={"task": URIRef(task_iri)},
    )


def get_first_query_result_if_exists(query_method: Callable, *args) -> Optional[str]:
    query_result = next(
        iter(list(query_method(*args))),
        None,
    )

    if query_result is None:
        return None

    return query_result


def get_data_properties_by_entity_iri(entity_iri: str, kg: Graph) -> query.Result:
    return kg.query(
        "\nSELECT ?p ?r WHERE {?p rdfs:domain ?entity_iri . "
        "?p rdfs:range ?r . "
        "?p rdf:type owl:DatatypeProperty . }",
        initBindings={"entity_iri": URIRef(entity_iri)},
    )


def get_method_properties_and_methods(input_kg, namespace_prefix, entity_parent_iri: str) -> query.Result:
    return input_kg.query(
        "\nSELECT ?p ?m WHERE {?p rdfs:domain ?entity_iri . "
        "?p rdfs:range ?m . "
        "?m rdfs:subClassOf " + namespace_prefix + ":AtomicMethod . }",
        initBindings={"entity_iri": URIRef(entity_parent_iri)},
    )


def get_input_properties_and_inputs(input_kg, namespace_prefix, entity_parent_iri: str) -> query.Result:
    return input_kg.query(
        "\nSELECT ?p ?m WHERE {?p rdfs:domain ?entity_iri . "
        "?p rdfs:range ?m ."
        "?p rdfs:subPropertyOf " + namespace_prefix + ":hasInput ."
        "?m rdfs:subClassOf " + namespace_prefix + ":DataEntity . }",
        initBindings={"entity_iri": URIRef(entity_parent_iri)},
    )


def get_output_properties_and_outputs(input_kg, namespace_prefix, entity_parent_iri: str) -> query.Result:
    return input_kg.query(
        "\nSELECT ?p ?m WHERE {?p rdfs:domain ?entity_iri . "
        "?p rdfs:range ?m ."
        "?p rdfs:subPropertyOf " + namespace_prefix + ":hasOutput ."
        "?m rdfs:subClassOf " + namespace_prefix + ":DataEntity . }",
        initBindings={"entity_iri": URIRef(entity_parent_iri)},
    )


def query_pipeline_info(kg, namespace_prefix):
    return kg.query(
        f"\nSELECT ?p ?i ?t WHERE {{?p rdf:type {namespace_prefix}:Pipeline ;"
        f"                          {namespace_prefix}:hasInputDataPath ?i ;"
        f"                          {namespace_prefix}:hasStartTask ?t . }}"
    )


def get_subclasses_of(class_iri: str, kg: Graph) -> query.Result:
    return kg.query(
        "\nSELECT ?t WHERE {?t rdfs:subClassOf ?class_iri . }",
        initBindings={"class_iri": class_iri},
    )


def get_data_properties_plus_inherited_by_class_iri(kg: Graph, entity_iri: str) -> List:
    """
    Retrieves data properties plus the inherited ones, given an entity IRI
    Args:
        kg: Graph object to use when querying
        entity_iri: IRI of entity to query

    Returns:
        List: contains rows of data property IRIs and their range
    """
    property_list = list(get_data_properties_by_entity_iri(entity_iri, kg))
    method_parent_classes = list(query_method_parent_classes(kg, entity_iri))
    for method_class_result_row in method_parent_classes:
        property_list += list(get_data_properties_by_entity_iri(method_class_result_row[0], kg))

    return property_list


def get_pipeline_and_first_task_iri(kg: Graph, namespace_prefix: str) -> Tuple[str, str, str]:
    """
    Retrieves the necessary information needed to start parsing a pipeline
    Args:
        kg: Graph object to use when querying
        namespace_prefix: namespace prefix to use when querying

    Returns:
        Tuple[str, str, str]: contains the pipeline IRI, the input data path and the first task's IRI
    """
    # assume one pipeline per file
    query_result = get_first_query_result_if_exists(
        query_pipeline_info,
        kg,
        namespace_prefix,
    )
    if query_result is None:
        print("Error: Pipeline info not found")
        exit(1)

    pipeline_iri, input_data_path, task_iri = query_result

    return str(pipeline_iri), str(input_data_path), str(task_iri)


def get_method_by_task_iri(
    kg: Graph,
    namespace_prefix: str,
    namespace: Namespace,
    task_iri: str,
) -> Optional[Entity]:
    """
    Retrieves a task's method, given a task IRI
    Args:
        kg: Graph object to use when querying
        namespace_prefix: namespace prefix to use when querying
        namespace: namespace to use when querying
        task_iri: IRI of task to query

    Returns:
        Optional[Entity]: object containing found method's basic info
                          is equal to None if method IRI wasn't found in KG
    """
    query_result = get_first_query_result_if_exists(
        query_method_iri_by_task_iri,
        kg,
        namespace_prefix,
        task_iri,
    )
    if query_result is None:
        return None

    method_iri = str(query_result[0])

    query_result = get_first_query_result_if_exists(
        query_entity_parent_iri,
        kg,
        method_iri,
        namespace.AtomicMethod,
    )
    if query_result is None:
        return None

    method_parent_iri = str(query_result[0])

    return Entity(method_iri, Entity(method_parent_iri))



from abc import abstractmethod

from classes.entity import Entity
from classes.task import Task
from utils.task_utils.visual_utils import *

"""
❗ Important for contributors: See the package's README.md before extending the code's functionality.
"""


class CanvasTaskCanvasMethod(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)
        self.grid = None
        self.fig = None
        self.has_layout = None
        self.has_canvas_name = None

    def run_method(self, *args):
        self.fig, self.grid = canvas_creation(self.has_layout)

        return None


class PlotTask(Task):
    def __init__(self, iri: str, parent_entity: Entity, canvas_method: CanvasTaskCanvasMethod):
        super().__init__(iri, parent_entity)
        self.fig = canvas_method.fig
        self.grid = canvas_method.grid

        self.has_legend_name = None
        self.has_layout = None
        self.has_y_label = None
        self.has_x_label = None
        self.has_y_lim = None
        self.has_x_lim = None
        self.has_line_style = None
        self.has_line_width = 1

    @abstractmethod
    def run_method(self, *args):
        raise NotImplementedError


class PlotTaskScatterplotMethod(PlotTask):
    def __init__(self, iri: str, parent_entity: Entity, canvas_method: CanvasTaskCanvasMethod):
        super().__init__(iri, parent_entity, canvas_method)
        self.has_scatter_size = None
        self.has_scatter_style = None

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame):
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_data = list(input_dict.values())[0]  # one input expected
        scatter_plot(
            data=input_data,
            fig=self.fig,
            grid=self.grid,
            layout=self.has_layout,
            line_width=self.has_line_width,
            line_style=self.has_line_style,
            legend_name=self.has_legend_name,
            x_lim=self.has_x_lim,
            y_lim=self.has_y_lim,
            x_label=self.has_x_label,
            y_label=self.has_y_label,
        )

        return None


class PlotTaskLineplotMethod(PlotTask):
    def __init__(self, iri: str, parent_entity: Entity, canvas_method: CanvasTaskCanvasMethod):
        super().__init__(iri, parent_entity, canvas_method)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame):
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_data = list(input_dict.values())[0]  # one input expected
        line_plot(
            data=input_data,
            fig=self.fig,
            grid=self.grid,
            layout=self.has_layout,
            line_width=self.has_line_width,
            line_style=self.has_line_style,
            legend_name=self.has_legend_name,
            x_lim=self.has_x_lim,
            y_lim=self.has_y_lim,
            x_label=self.has_x_label,
            y_label=self.has_y_label,
        )

        return None



import os
from typing import Dict, Union

import pandas as pd
from rdflib import Literal
from utils.cli_utils import (get_input_for_existing_data_entities,
                             get_input_for_new_data_entities)
from utils.kg_creation_utils import (add_and_attach_data_entity,
                                     add_data_entity_instance,
                                     add_instance_from_parent_with_relation,
                                     add_literal, create_pipeline_task,
                                     name_instance)
from utils.query_utils import *
from utils.query_utils import (get_data_properties_plus_inherited_by_class_iri,
                               get_method_by_task_iri,
                               get_pipeline_and_first_task_iri)
from utils.string_utils import property_name_to_field_name

from .data_entity import DataEntity
from .entity import Entity
from .kg_schema import KGSchema
from .task import Task
from .tasks import ml_tasks, statistic_tasks, visual_tasks

KG_SCHEMAS = {
    "Data Science": {
        "path": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/ds_exeKGOntology.ttl",
        "namespace": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/ds_exeKGOntology.ttl#",
        "namespace_prefix": "ds",
    },
    "Visualization": {
        "path": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/visu_exeKGOntology.ttl",
        "namespace": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/visu_exeKGOntology.ttl#",
        "namespace_prefix": "visu",
    },
    "Statistics": {
        "path": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/stats_exeKGOntology.ttl",
        "namespace": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/stats_exeKGOntology.ttl#",
        "namespace_prefix": "stats",
    },
    "Machine Learning": {
        "path": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/ml_exeKGOntology.ttl",
        "namespace": "https://raw.githubusercontent.com/nsai-uio/ExeKGOntology/main/ml_exeKGOntology.ttl#",
        "namespace_prefix": "ml",
    },
}


class ExeKG:
    def __init__(self, kg_schema_name: str = None, input_exe_kg_path: str = None):
        """

        Args:
            kg_schema_name: name of chosen bottom-level KG schema to use in case of KG construction (must be equal to one of KG_SCHEMAS keys)
                            acts as switch for KG construction mode (if filled, mode is on)
            input_exe_kg_path: path of KG to be executed
                               acts as switch for KG execution mode (if filled, mode is on)
        """
        self.top_level_schema = KGSchema.from_schema_info(KG_SCHEMAS["Data Science"])  # top-level KG schema

        # top-level KG schema entities
        self.atomic_task = Entity(self.top_level_schema.namespace.AtomicTask)
        self.atomic_method = Entity(self.top_level_schema.namespace.AtomicMethod)
        self.data_entity = Entity(self.top_level_schema.namespace.DataEntity)
        self.pipeline = Entity(self.top_level_schema.namespace.Pipeline)
        self.data = Entity(self.top_level_schema.namespace.Data)
        self.data_semantics = Entity(self.top_level_schema.namespace.DataSemantics)
        self.data_structure = Entity(self.top_level_schema.namespace.DataStructure)

        # self.input_kg: KG eventually filled with 3 KG schemas and the input executable KG in case of KG execution
        self.input_kg = Graph(bind_namespaces="rdflib")
        if input_exe_kg_path:  # KG execution mode
            self.input_kg.parse(input_exe_kg_path, format="n3")  # parse input executable KG
            all_ns = [n for n in self.input_kg.namespace_manager.namespaces()]
            bottom_level_schema_info_set = False  # flag indicating that a bottom-level schema was found
            for schema_name, schema_info in KG_SCHEMAS.items():  # search for used bottom-level schema
                if (
                    schema_name == "Data Science" or schema_name == "Visualization"
                ):  # skip top-level KG schema and Visualization schema that is always used
                    continue

                if (schema_info["namespace_prefix"], URIRef(schema_info["namespace"])) in all_ns:
                    # bottom-level schema found
                    self.bottom_level_schema = KGSchema.from_schema_info(schema_info)
                    bottom_level_schema_info_set = True
                    break

            visu_schema_info = KG_SCHEMAS["Visualization"]
            if (
                not bottom_level_schema_info_set
                and (visu_schema_info["namespace_prefix"], URIRef(visu_schema_info["namespace"])) in all_ns
            ):  # Visualization schema is considered the bottom-level schema ONLY IF no other bottom-level schema was found
                self.bottom_level_schema = KGSchema.from_schema_info(visu_schema_info)
                bottom_level_schema_info_set = True

            if not bottom_level_schema_info_set:  # no bottom-level schema found, input executable KG is invalid
                print("Input executable KG did not have any bottom level KG schemas")
                exit(1)
        else:  # KG construction mode
            # bottom-level schema used as compatibility guide for constructing executable KG
            self.bottom_level_schema = KGSchema.from_schema_info(KG_SCHEMAS[kg_schema_name])

        self.visu_schema = KGSchema.from_schema_info(
            KG_SCHEMAS["Visualization"]
        )  # Visualization KG schema, always used

        self.input_kg += (
            self.top_level_schema.kg + self.bottom_level_schema.kg + self.visu_schema.kg
        )  # combine all KG schemas in input KG

        self.output_kg = Graph(bind_namespaces="rdflib")  # KG to be filled while constructing executable KG

        self._bind_used_namespaces([self.input_kg, self.output_kg])

        # below variables are filled in self._parse_kgs()
        self.task_type_dict = {}  # dict for uniquely naming each new pipeline task
        self.method_type_dict = {}  # dict for uniquely naming each new pipeline method
        self.atomic_task_list = []  # list for storing the available sub-classes of ds:AtomicTask
        self.atomic_method_list = []  # list for storing the available sub-classes of ds:AtomicMethod
        self.data_type_list = []  # list for storing the available sub-classes of ds:DataEntity
        self.data_semantics_list = []  # list for storing the available sub-classes of ds:DataSemantics
        self.data_structure_list = []  # list for storing the available sub-classes of ds:DataStructure

        self.existing_data_entity_list = (
            []
        )  # contains existing data entities that are output entities of previous tasks during KG construction
        self.last_created_task = (
            None  # last created pipeline task, for connecting consecutive pipeline tasks during KG construction
        )
        self.canvas_task_created = False  # indicates if canvas task was created during KG construction, and used for hiding the other Visualization tasks in CLI

        self._parse_kgs()

    def _bind_used_namespaces(self, kgs: List[Graph]):
        """
        Binds top-level, bottom-level and Visualization KG schemas' namespaces with their prefixes
        Adds these bindings to the Graphs of kgs list
        Args:
            kgs: list of Graph objects to which the namespace bindings are added
        """
        for kg in kgs:
            kg.bind(self.top_level_schema.namespace_prefix, self.top_level_schema.namespace)
            kg.bind(
                self.bottom_level_schema.namespace_prefix,
                self.bottom_level_schema.namespace,
            )
            kg.bind(
                self.visu_schema.namespace_prefix,
                self.visu_schema.namespace,
            )

    def _parse_kgs(self) -> None:
        """
        Fills lists with subclasses of top-level KG schema classes and initializes dicts used for unique naming
        """
        atomic_task_subclasses = get_subclasses_of(self.atomic_task.iri, self.input_kg)
        for t in list(atomic_task_subclasses):
            task = Entity(t[0], self.atomic_task)
            self.atomic_task_list.append(task)
            self.task_type_dict[task.name] = 1

        atomic_method_subclasses = get_subclasses_of(self.atomic_method.iri, self.input_kg)
        for m in list(atomic_method_subclasses):
            method = Entity(m[0], self.atomic_method)
            self.atomic_method_list.append(method)
            self.method_type_dict[method.name] = 1

        data_type_subclasses = get_subclasses_of(self.data_entity.iri, self.input_kg)
        for d in list(data_type_subclasses):
            data_type = Entity(d[0], self.data_entity)
            self.data_type_list.append(data_type)

        data_semantics_subclasses = get_subclasses_of(self.data_semantics.iri, self.top_level_schema.kg)
        for d in list(data_semantics_subclasses):
            if d[0] == self.data_entity.iri:
                continue
            data_semantics = Entity(d[0], self.data_semantics)
            self.data_semantics_list.append(data_semantics)

        data_structure_subclasses = get_subclasses_of(self.data_structure.iri, self.top_level_schema.kg)
        for d in list(data_structure_subclasses):
            if d[0] == self.data_entity.iri:
                continue
            data_structure = Entity(d[0], self.data_structure)
            self.data_structure_list.append(data_structure)

    def create_pipeline_task(self, pipeline_name: str, input_data_path: str) -> Task:
        """
        Instantiates and adds a new pipeline task entity to self.output_kg
        Args:
            pipeline_name: name for the pipeline
            input_data_path: path for the input data to be used by the pipeline's tasks

        Returns:
            Task: created pipeline
        """
        pipeline = create_pipeline_task(
            self.top_level_schema.namespace,
            self.bottom_level_schema.namespace,
            self.pipeline,
            self.output_kg,
            pipeline_name,
            input_data_path,
        )
        self.last_created_task = pipeline
        return pipeline

    def create_data_entity(
        self,
        name: str,
        source_value: str,
        data_semantics_name: str,
        data_structure_name: str,
    ) -> DataEntity:
        """
        Creates a DataEntity object
        Args:
            name: name of the data entity
            source_value: name of the data source corresponding to a column of the data
            data_semantics_name: name of the data semantics entity
            data_structure_name: name of the data structure entity

        Returns:
            DataEntity: object initialized with the given parameter values
        """
        return DataEntity(
            self.bottom_level_schema.namespace + name,
            self.data_entity,
            source_value,
            self.top_level_schema.namespace + data_semantics_name,
            self.top_level_schema.namespace + data_structure_name,
        )

    def add_task(
        self,
        task_type: str,
        input_data_entity_dict: Dict[str, List[DataEntity]],
        method_type: str,
        data_properties: Dict[str, Union[str, int, float]],
        visualization: bool = False,
    ) -> Task:
        """
        Instantiates and adds a new task entity to self.output_kg
        Components added to the task during creation: input and output entities, and a method with data properties
        Args:
            task_type: type of the task
            input_data_entity_dict: keys -> input entity names corresponding to the given task_type as defined in the chosen bottom-level KG schema
                                    values -> list of corresponding data entities to be added as input to the task
            method_type: type of the task's method
            data_properties: keys -> data property names corresponding to the given method_type as defined in the chosen bottom-level KG schema
                             values -> list of corresponding values to be added as property values to the task
            visualization: if True, the namespace prefix of Visualization KG schema is used during creation of the task
                           else, the namespace prefix of the chosen bottom-level KG schema is used

        Returns:
            Task: object of the created task
        """
        namespace_to_use = (
            self.visu_schema.namespace if visualization else self.bottom_level_schema.namespace
        )  # use appropriate namespace for the task

        relation_iri = (
            self.top_level_schema.namespace.hasNextTask
            if self.last_created_task.type != "Pipeline"
            else self.top_level_schema.namespace.hasStartTask
        )  # use relation depending on the previous task

        # instantiate task and link it with the previous one
        parent_task = Task(namespace_to_use + task_type, self.atomic_task)
        added_entity = add_instance_from_parent_with_relation(
            self.bottom_level_schema.namespace,
            self.output_kg,
            parent_task,
            relation_iri,
            self.last_created_task,
            name_instance(self.task_type_dict, self.method_type_dict, parent_task),
        )
        next_task = Task.from_entity(added_entity)  # create Task object from Entity object

        # instantiate and add given input data entities to the task
        self._add_inputs_to_task(next_task, input_data_entity_dict)
        # instantiate and add output data entities to the task, as specified in the KG schema
        self._add_outputs_to_task(next_task)

        method_parent = Entity(namespace_to_use + method_type, self.atomic_method)

        # fetch compatible methods and their properties from KG schema
        results = list(
            get_method_properties_and_methods(
                self.input_kg,
                self.top_level_schema.namespace_prefix,
                next_task.parent_entity.iri,
            )
        )

        chosen_property_method = next(
            filter(lambda pair: pair[1].split("#")[1] == method_type, results), None
        )  # match given method_type with query result
        if chosen_property_method is None:
            print(f"Property connecting task of type {task_type} with method of type {method_type} not found")
            exit(1)

        # instantiate method and link it with the task using the appropriate chosen_property_method[0] relation
        add_instance_from_parent_with_relation(
            self.bottom_level_schema.namespace,
            self.output_kg,
            method_parent,
            chosen_property_method[0],
            next_task,
            name_instance(self.task_type_dict, self.method_type_dict, method_parent),
        )

        # fetch compatible data properties from KG schema
        property_list = get_data_properties_plus_inherited_by_class_iri(self.input_kg, method_parent.iri)

        # add data properties to the task with given values
        for pair in property_list:
            property_iri = pair[0]
            property_name = property_iri.split("#")[1]
            range_iri = pair[1]
            input_property = Literal(
                lexical_or_value=data_properties[property_name],
                datatype=range_iri,
            )
            add_literal(self.output_kg, next_task, property_iri, input_property)

        self.last_created_task = next_task  # store created task

        return next_task

    def _add_inputs_to_task(
        self, task_entity: Task, input_data_entity_dict: Dict[str, List[DataEntity]] = None
    ) -> None:
        """
        Instantiates and adds given input data entities to the given task of self.output_kg
        if input_data_entity_dict is None, user is asked to specify input data entities
        Args:
            task_entity: the task to add the input to
            input_data_entity_dict: keys -> input entity names corresponding to the given task as defined in the chosen bottom-level KG schema
                                    values -> list of corresponding data entities to be added as input to the task
        """
        # fetch compatible inputs from KG schema
        results = list(
            get_input_properties_and_inputs(
                self.input_kg,
                self.top_level_schema.namespace_prefix,
                task_entity.parent_entity.iri,
            )
        )

        # task_type_index was incremented when creating the task entity
        # reset the index to match the currently created task's index
        task_type_index = self.task_type_dict[task_entity.type] - 1
        for _, input_entity_iri in results:
            input_entity_name = input_entity_iri.split("#")[1]
            if input_data_entity_dict:
                input_data_entity_list = input_data_entity_dict[input_entity_name]
            else:
                # use CLI
                print(f"Specify input corresponding to {input_entity_name}")
                input_data_entity_list = get_input_for_existing_data_entities(self.existing_data_entity_list)
                input_data_entity_list += get_input_for_new_data_entities(
                    self.data_semantics_list,
                    self.data_structure_list,
                    self.bottom_level_schema.namespace,
                    self.data_entity,
                )

            same_input_index = 1
            for input_data_entity in input_data_entity_list:
                # instantiate data entity corresponding to the found input_entity_name
                data_entity_iri = input_entity_iri + str(task_type_index) + "_" + str(same_input_index)
                # instantiate given data entity
                add_data_entity_instance(
                    self.output_kg,
                    self.data,
                    self.top_level_schema.kg,
                    self.top_level_schema.namespace,
                    input_data_entity,
                )
                # instantiate and attach data entity with reference to the given data entity
                data_entity = DataEntity(
                    data_entity_iri,
                    DataEntity(input_entity_iri, self.data_entity),
                    has_reference=input_data_entity.iri,
                )
                add_and_attach_data_entity(
                    self.output_kg,
                    self.data,
                    self.top_level_schema.kg,
                    self.top_level_schema.namespace,
                    data_entity,
                    self.top_level_schema.namespace.hasInput,
                    task_entity,
                )
                task_entity.input_dict[input_entity_name] = data_entity
                same_input_index += 1

    def _add_outputs_to_task(self, task_entity: Task) -> None:
        """
        Instantiates and adds output data entities to the given task of self.output_kg, based on the task's definition in the KG schema
        Args:
            task_entity: the task to add the output to
        """
        # fetch compatible outputs from KG schema
        results = list(
            get_output_properties_and_outputs(
                self.input_kg,
                self.top_level_schema.namespace_prefix,
                task_entity.parent_entity.iri,
            )
        )

        # task_type_index was incremented when creating the task entity
        # reset the index to match the currently created task's index
        task_type_index = self.task_type_dict[task_entity.type] - 1
        for output_property, output_entity_iri in results:
            # instantiate and add data entity
            data_entity_iri = output_entity_iri + str(task_type_index)
            data_entity = DataEntity(data_entity_iri, self.data_entity)
            add_and_attach_data_entity(
                self.output_kg,
                self.data,
                self.top_level_schema.kg,
                self.top_level_schema.namespace,
                data_entity,
                self.top_level_schema.namespace.hasOutput,
                task_entity,
            )
            task_entity.output_dict[output_entity_iri.split("#")[1]] = data_entity
            self.existing_data_entity_list.append(data_entity)

    def _create_next_task_cli(self) -> Union[None, Task]:
        """
        Instantiates and adds task (without method) based on user input to self.output_kg
        Adds task's output data entities to self.existing_data_entity_list
        Returns:
            None: in case user wants to end the pipeline creation
            Task: object of the created task
        """
        print("Please choose the next task")
        for i, t in enumerate(self.atomic_task_list):
            if not self.canvas_task_created and t.name == "PlotTask":
                continue
            if self.canvas_task_created and t.name == "CanvasTask":
                continue
            print(f"\t{str(i)}. {t.name}")
        print(f"\t{str(-1)}. End pipeline")
        next_task_id = int(input())
        if next_task_id == -1:
            return None

        next_task_parent = self.atomic_task_list[next_task_id]
        relation_iri = (
            self.top_level_schema.namespace.hasNextTask
            if self.last_created_task.type != "Pipeline"
            else self.top_level_schema.namespace.hasStartTask
        )  # use relation depending on the previous task

        # instantiate task and link it with the previous one
        task_entity = add_instance_from_parent_with_relation(
            self.bottom_level_schema.namespace,
            self.output_kg,
            next_task_parent,
            relation_iri,
            self.last_created_task,
            name_instance(self.task_type_dict, self.method_type_dict, next_task_parent),
        )

        task_entity = Task(task_entity.iri, task_entity.parent_entity)  # create Task object from Entity object's info

        # instantiate and add input data entities to the task based on user input
        self._add_inputs_to_task(task_entity)
        # instantiate and add output data entities to the task, as specified in the KG schema
        self._add_outputs_to_task(task_entity)

        self.last_created_task = task_entity
        if task_entity.type == "CanvasTask":
            self.canvas_task_created = True

        return task_entity

    def _create_method(self, task_to_attach_to: Entity) -> None:
        """
        Instantiate and attach method to task of self.output_kg
        Args:
            task_to_attach_to: the task to attach the created method to
        """
        print(f"Please choose a method for {task_to_attach_to.type}:")

        # fetch compatible methods and their properties from KG schema
        results = list(
            get_method_properties_and_methods(
                self.input_kg,
                self.top_level_schema.namespace_prefix,
                task_to_attach_to.parent_entity.iri,
            )
        )
        for i, pair in enumerate(results):
            tmp_method = pair[1].split("#")[1]
            print(f"\t{str(i)}. {tmp_method}")

        method_id = int(input())
        selected_property_and_method = results[method_id]
        method_parent = next(
            filter(
                lambda m: m.iri == selected_property_and_method[1],
                self.atomic_method_list,
            ),
            None,
        )
        # instantiate method and link it with the task using the appropriate selected_property_and_method[0] relation
        add_instance_from_parent_with_relation(
            self.bottom_level_schema.namespace,
            self.output_kg,
            method_parent,
            selected_property_and_method[0],
            task_to_attach_to,
            name_instance(self.task_type_dict, self.method_type_dict, method_parent),
        )

        # fetch compatible data properties from KG schema
        property_list = get_data_properties_plus_inherited_by_class_iri(self.input_kg, method_parent.iri)

        if property_list:
            print(f"Please enter requested properties for {method_parent.name}:")
            # add data properties to the task with given values
            for pair in property_list:
                property_instance = URIRef(pair[0])
                range = pair[1].split("#")[1]
                range_iri = pair[1]
                input_property = Literal(
                    lexical_or_value=input("\t{} in range({}): ".format(pair[0].split("#")[1], range)),
                    datatype=range_iri,
                )
                add_literal(self.output_kg, task_to_attach_to, property_instance, input_property)

    def start_pipeline_creation(self, pipeline_name: str, input_data_path: str) -> None:
        """
        Handles the pipeline creation through CLI
        Args:
            pipeline_name: name for the pipeline
            input_data_path: path for the input data to be used by the pipeline's tasks
        """
        pipeline = create_pipeline_task(
            self.top_level_schema.namespace,
            self.bottom_level_schema.namespace,
            self.pipeline,
            self.output_kg,
            pipeline_name,
            input_data_path,
        )

        self.last_created_task = pipeline

        while True:
            next_task = self._create_next_task_cli()
            if next_task is None:
                break

            self._create_method(next_task)

    def save_created_kg(self, file_path: str) -> None:
        """
        Saves self.output_kg to a file
        Args:
            file_path: path of the output file
        """
        dir_path = os.path.dirname(file_path)
        os.makedirs(dir_path, exist_ok=True)

        self.output_kg.serialize(destination=file_path)
        print(f"Executable KG saved in {file_path}")

    def _property_value_to_field_value(self, property_value: str) -> Union[str, DataEntity]:
        """
        Converts property value to Python class field value
        If property_value is not a data entity's IRI, it is returned as is
        Else, its property values are converted recursively and stored in a DataEntity object
        Args:
            property_value: value of the property as found in KG

        Returns:
            str: property_value parameter as is
            DataEntity: object containing parsed data entity properties
        """
        if "#" in property_value:
            data_entity = self._parse_data_entity_by_iri(property_value)
            if data_entity is None:
                return property_value
            return data_entity

        return property_value

    def _parse_data_entity_by_iri(self, in_out_data_entity_iri: str) -> Optional[DataEntity]:
        """
        Parses an input or output data entity of self.input_kg and stores the parsed info in a Python object
        Args:
            in_out_data_entity_iri: IRI of the KG entity to parse

        Returns:
            None: if given IRI does not belong to an instance of a sub-class of self.top_level_schema.namespace.DataEntity
            DataEntity: object with data entity's parsed properties
        """
        # fetch type of entity with given IRI
        query_result = get_first_query_result_if_exists(
            query_entity_parent_iri,
            self.input_kg,
            in_out_data_entity_iri,
            self.top_level_schema.namespace.DataEntity,
        )
        if query_result is None:
            return None

        data_entity_parent_iri = str(query_result[0])

        # fetch IRI of data entity that is referenced by the given entity
        query_result = get_first_query_result_if_exists(
            query_data_entity_reference_iri,
            self.input_kg,
            self.top_level_schema.namespace_prefix,
            in_out_data_entity_iri,
        )

        if query_result is None:  # no referenced data entity found
            data_entity_ref_iri = in_out_data_entity_iri
        else:
            data_entity_ref_iri = str(query_result[0])

        # create DataEntity object to store all the parsed properties
        data_entity = DataEntity(in_out_data_entity_iri, Entity(data_entity_parent_iri))
        data_entity.has_reference = data_entity_ref_iri.split("#")[1]

        for s, p, o in self.input_kg.triples((URIRef(data_entity_ref_iri), None, None)):
            # parse property name and value
            field_name = property_name_to_field_name(str(p))
            if not hasattr(data_entity, field_name) or field_name == "type":
                continue
            field_value = self._property_value_to_field_value(str(o))
            setattr(data_entity, field_name, field_value)  # set field value dynamically

        return data_entity

    def _parse_task_by_iri(self, task_iri: str, canvas_method: visual_tasks.CanvasTaskCanvasMethod = None) -> Task:
        """
        Parses a task of self.input_kg and stores the info in an object of a sub-class of Task
        The sub-class name and the object's fields are mapped dynamically based on the found KG components
        Args:
            task_iri: IRI of the task to be parsed
            canvas_method: optional object to pass as argument for task object initialization

        Returns:
            Task: object of a sub-class of Task, containing all the parsed info
        """
        # fetch type of entity with given IRI
        query_result = get_first_query_result_if_exists(
            query_entity_parent_iri,
            self.input_kg,
            task_iri,
            self.top_level_schema.namespace.AtomicTask,
        )

        if (
            query_result is None
        ):  # given IRI does not belong to an instance of a sub-class of self.top_level_schema.namespace.AtomicTask
            print(f"Cannot retrieve parent of task with iri {task_iri}. Exiting...")
            exit(1)

        task_parent_iri = str(query_result[0])

        task = Task(task_iri, Task(task_parent_iri))
        method = get_method_by_task_iri(
            self.input_kg,
            self.top_level_schema.namespace_prefix,
            self.top_level_schema.namespace,
            task_iri,
        )
        if method is None:
            print(f"Cannot retrieve method for task with iri: {task_iri}")

        # perform automatic mapping of KG task class to Python sub-class
        class_name = task.type + method.type
        Class = getattr(visual_tasks, class_name, None)
        if Class is None:
            Class = getattr(statistic_tasks, class_name, None)
        if Class is None:
            Class = getattr(ml_tasks, class_name, None)

        # create Task sub-class object
        if canvas_method:
            task = Class(task_iri, Task(task_parent_iri), canvas_method)
        else:
            task = Class(task_iri, Task(task_parent_iri))

        for s, p, o in self.input_kg.triples((URIRef(task_iri), None, None)):
            # parse property name and value
            field_name = property_name_to_field_name(str(p))
            if not hasattr(task, field_name) or field_name == "type":
                continue
            field_value = self._property_value_to_field_value(str(o))

            # set field value dynamically
            if field_name == "has_input" or field_name == "has_output":
                getattr(task, field_name).append(field_value)
            else:
                setattr(task, field_name, field_value)

        return task

    def execute_pipeline(self):
        """
        Retrieves and executes pipeline by parsing self.input_kg
        """
        pipeline_iri, input_data_path, next_task_iri = get_pipeline_and_first_task_iri(
            self.input_kg, self.top_level_schema.namespace_prefix
        )
        input_data = pd.read_csv(input_data_path, delimiter=",", encoding="ISO-8859-1")
        canvas_method = None  # stores Task object that corresponds to a task of type CanvasTask
        task_output_dict = {}  # gradually filled with outputs of executed tasks
        while next_task_iri is not None:
            next_task = self._parse_task_by_iri(next_task_iri, canvas_method)
            output = next_task.run_method(task_output_dict, input_data)
            if output:
                task_output_dict.update(output)

            if next_task.type == "CanvasTask":
                canvas_method = next_task

            next_task_iri = next_task.has_next_task

    @staticmethod
    def input_kg_schema_name() -> str:
        """
        Prompts the user to choose a schema by presenting the available schemas' names
        Returns:
            str: chosen schema name
        """
        kg_schema_names = list(KG_SCHEMAS.keys())
        print(
            "Choose a KG schema to use. Components of the Visualization schema can be used regardless of the chosen schema."
        )
        for i, kg_schema_name in enumerate(kg_schema_names):
            if kg_schema_name == "Data Science":
                continue
            print(f"{i}: {kg_schema_name}")
        selected_schema_i = int(input())
        selected_schema_name = kg_schema_names[selected_schema_i]

        return selected_schema_name



from abc import abstractmethod
from typing import Dict

import numpy as np
import pandas as pd

from .entity import Entity


class Task(Entity):
    """
    Abstraction of owl:class Task.

    ❗ Important for contributors ❗
    The fields that contain "_" are by convention the snake-case conversions of the equivalent camel-case property names in the KG.
    e.g. has_next_task field corresponds to hasNextTask property in the KG.
    This is necessary for automatically mapping KG properties to Python object fields while parsing the KG.
    """

    def __init__(
        self,
        iri: str,
        parent_entity: Entity = None,
    ):
        super().__init__(iri, parent_entity)
        self.has_next_task = None
        self.has_method = None
        self.has_input = []
        self.has_output = []
        self.input_dict = {}  # used for storing input DataEntity objects during KG creation
        self.output_dict = {}  # used for storing output DataEntity objects during KG creation

    @classmethod
    def from_entity(cls, entity: Entity):
        return cls(entity.iri, entity.parent_entity)

    def create_output_dict(self, keyword_value_dict: dict) -> dict:
        """
        For each key in keyword_value_dict, checks if the key exists in an output name of the Task.
        If yes, adds the output name with its value to out_dict.
        Args:
            keyword_value_dict: key-value pairs where key is a keyword to find in an output name of the Task
                                  and value is the value corresponding to that output name

        Returns:
            dict: pairs of Task's output names and corresponding output values
        """
        if len(self.has_output) == 0:
            # assume one output and use task name as key
            return {self.name: list(keyword_value_dict.values())[0]}

        output_names = [has_output_elem.name for has_output_elem in self.has_output]
        out_dict = {}
        for output_name in output_names:
            for key, value in keyword_value_dict.items():
                if key in output_name:
                    out_dict[output_name] = value

        return out_dict

    def get_inputs(self, dict_to_search: dict, fallback_df: pd.DataFrame) -> Dict[str, np.ndarray]:
        """
        Tries to match the Task's input names with the keys of dict_to_search
        and fills input_dict list with their corresponding values.
        If the matches fail, it retrieves columns of the provided fallback_df
        Args:
            dict_to_search: contains key-value pairs where key is a possible input name and value is its corresponding value
            fallback_df: contains data to return as an alternative

        Returns:
            Dict[str, np.ndarray]: pairs of input entity types and corresponding input values
        """
        input_dict = {}
        for input in self.has_input:
            try:
                input_dict[input.type] = dict_to_search[input.has_reference]
            except KeyError:
                input_dict[input.type] = fallback_df[input.has_source]

        return input_dict

    @abstractmethod
    def run_method(self, *args):
        """
        Abstract method to be implemented by Task sub-classes that are in the bottom of the hierarchy.
        Executes the logic that is needed to fulfill the Task.
        Args:
            *args: defined by sub-classes
        """
        raise NotImplementedError



import numpy as np
import pandas as pd


def trend_calculation(
    input_data: np.ndarray,
    half_window_size: int = 2,
    padding: str = "same",
) -> np.ndarray:
    """calculate the trend of the data, which is the sliding-window average"""

    def padding_input(input: np.ndarray, half_window_size: int, padding: str = "same") -> np.ndarray:
        """padding the beginning and end of the input data by the beginning or end value"""
        if padding == "same":
            begin_padding = pd.Series([input.iloc[0]] * half_window_size)
            end_padding = pd.Series([input.iloc[len(input) - 1]] * half_window_size)
            output = pd.concat([begin_padding, input, end_padding]).reset_index().drop("index", axis=1)
            return output

        else:
            return 0

    input = padding_input(input_data, half_window_size)

    output = 0
    input_len = len(input) - 2 * half_window_size
    for i in range(2 * half_window_size + 1):
        output += np.array(input.iloc[i : i + input_len])

    output = output / (2 * half_window_size + 1)

    return output


def mean_calculation(input_data: np.ndarray) -> np.ndarray:
    return np.mean(input_data)


def std_deviation(input_data: np.ndarray) -> np.ndarray:
    return np.std(input_data)


def minimum_calculation(input_data: np.ndarray) -> np.ndarray:
    return np.min(input_data)


def maximum_calculation(input_data: np.ndarray) -> np.ndarray:
    return np.max(input_data)


def iqr_calculation(input_data: np.ndarray, percent: int = 50) -> np.ndarray:
    """return quarter of the input_data"""
    return np.percentile(input_data, percent)


def outlier_calculation(input: np.ndarray, iq1: float = None, iq3: float = None) -> np.ndarray:
    """return the outliers in the data"""
    iq1 = iqr_calculation(input, 25) if (not iq1) else iq1
    iq3 = iqr_calculation(input, 75) if (not iq3) else iq3
    median = np.median(input)
    iqr = iq3 - iq1
    high_outliers = input < median - 1.5 * iqr
    low_outliers = input > median + 1.5 * iqr
    outlier_rows = [low_outliers.iloc[i] or high_outliers.iloc[i] for i in range(len(input))]
    return outlier_rows


def scattering_calculation(input_data: np.ndarray) -> np.ndarray:
    trend_calc_output = trend_calculation(input_data)
    output = input_data - trend_calc_output

    return output


def normalization(input_data: np.ndarray) -> np.ndarray:
    mean = mean_calculation(input_data)
    std = std_deviation(input_data)
    output = (input_data - mean) / std

    return output



from typing import Optional, Tuple

import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.font_manager import FontProperties


def canvas_creation(layout: str) -> Tuple[Figure, Optional[plt.GridSpec]]:
    """the canvas task taking as input the starting task POs and output the fig and grid"""

    font = FontProperties()
    font.set_size(10)
    font.set_name("Verdana")

    try:
        n_rows, n_cols = (int(i) for i in layout[0].split(" "))
    except:
        n_rows, n_cols = (1, 1)

    fig = plt.figure(figsize=(7, 5))

    grid = None if (n_rows == n_cols and n_rows == 1) else plt.GridSpec(n_rows, n_cols, hspace=0.3, wspace=0.3)

    return fig, grid


# def scatter_plot(data: pd.Series, fig: Figure, grid: plt.GridSpec, layout: str, line_width: str, line_style: str,
#                  legend_name: str, x_lim: str, y_lim: str, x_label: str, y_label: str):
def scatter_plot(**kwargs):
    plot_creation("scatter", **kwargs)


def line_plot(**kwargs):
    plot_creation("line", **kwargs)


def plot_creation(
    plot_type: str,
    data: pd.Series,
    fig: Figure,
    grid: plt.GridSpec,
    layout: str,
    line_width: str,
    line_style: str,
    legend_name: str,
    x_lim: str,
    y_lim: str,
    x_label: str,
    y_label: str,
):
    plot = plt
    if grid is not None:
        row_start, row_end, col_start, col_end = (int(i) for i in layout.split())
        plot = fig.add_subplot(grid[row_start:row_end, col_start:col_end])

    if plot_type == "line":
        plot.plot(
            data.index,
            data,
            linestyle=line_style,
            linewidth=int(line_width),
            label=legend_name,
        )
    elif plot_type == "scatter":
        plot.scatter(
            data.index,
            data,
            s=int(int(line_width) * 10),
            marker=line_style,
            linewidth=int(line_width),
            label=legend_name,
        )
    else:
        print("Invalid plot type given")

    if grid is None:
        plot.xlim(x_lim)
        plot.ylim(y_lim)
        plot.xlabel(x_label)
        plot.ylabel(y_label)
    else:
        plot.set_xlim(x_lim)
        plot.set_ylim(y_lim)
        plot.set_xlabel(x_label)
        plot.set_ylabel(y_label)

    plt.legend()
    plt.show()



from typing import List, Tuple

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.neighbors import KNeighborsRegressor
from sklearn.neural_network import MLPRegressor


def concatenation(inputs: List[np.ndarray]) -> pd.DataFrame:
    return pd.concat(inputs, axis=1)


def data_splitting(
    input_x: pd.DataFrame, input_y: np.ndarray, split_ratio: str
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """split data into training and testing set"""

    splitting_point = int(float(split_ratio) * float(input_x.shape[0]))

    train_x = input_x.iloc[:splitting_point]
    test_x = input_x.iloc[splitting_point:]
    train_y = input_y.iloc[:splitting_point]
    test_y = input_y.iloc[splitting_point:]

    return train_x, test_x, train_y, test_y


def k_nn_train(
    input_x: np.ndarray, input_y: np.ndarray, n_neighbors: int = 3
) -> Tuple[KNeighborsRegressor, np.ndarray]:
    print("n_neighbors = ", n_neighbors)
    model = KNeighborsRegressor(n_neighbors=n_neighbors)
    model.fit(input_x, input_y)

    predicted_y = model.predict(input_x)

    print("KNN training finished")

    return model, predicted_y


def k_nn_test(model: KNeighborsRegressor, input_x: np.ndarray) -> np.ndarray:
    predicted_y = model.predict(input_x)

    print("KNN testing finished")

    return predicted_y


def lr_training(input_x: np.ndarray, input_y: np.ndarray) -> Tuple[LinearRegression, np.ndarray]:
    model = LinearRegression()
    model.fit(input_x, input_y)
    predicted_y = model.predict(input_x)

    print("LR training finished")
    return model, predicted_y


def lr_testing(model: LinearRegression, input_x: np.ndarray):
    predict_y = model.predict(input_x)

    return predict_y


def mlp_train(input_x: np.ndarray, input_y: np.ndarray, solver="adam") -> Tuple[MLPRegressor, np.ndarray]:
    model = MLPRegressor(solver=solver)
    model.fit(input_x, input_y)

    predicted_y = model.predict(input_x)

    print("MLP training finished")

    return model, predicted_y


def mlp_test(model: MLPRegressor, input_x: np.ndarray) -> np.ndarray:
    predicted_y = model.predict(input_x)

    print("MLP testing finished")

    return predicted_y


def ml_performance_calculation(
    real_train: np.ndarray = 0,
    real_test: np.ndarray = 0,
    predicted_train: np.ndarray = 0,
    predicted_test: np.ndarray = 0,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    train_err = pd.DataFrame(real_train - predicted_train)
    test_err = pd.DataFrame(real_test - predicted_test)

    return train_err, test_err



import re


def camel_to_snake(text: str) -> str:
    """
    Converts camel-case string to snake-case
    Args:
        text: string to convert

    Returns:
        str: converted string
    """
    text = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", text).lower()


def property_name_to_field_name(property_name: str) -> str:
    """
    Extracts property name from IRI and converts it to snake-case
    Args:
        property_name: IRI to parse

    Returns:
        str: converted string
    """
    return camel_to_snake(property_name.split("#")[1])



from classes.entity import Entity
from classes.task import Task
from utils.task_utils.statistic_utils import *

"""
❗ Important for contributors: See the package's README.md before extending the code's functionality.
"""


class TrendCalculationTaskTrendCalculationMethod(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_data = list(input_dict.values())[0]  # one input expected
        trend_calculation_result = trend_calculation(input_data)

        return self.create_output_dict({"DataOutTrendCalculation": trend_calculation_result})


class ScatteringCalculationTaskScatteringCalculationMethod(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_data = list(input_dict.values())[0]  # one input expected
        scattering_calculation_result = scattering_calculation(input_data)

        return self.create_output_dict({"DataOutScatteringCalculation": scattering_calculation_result})


class NormalizationTaskNormalizationMethod(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_data = list(input_dict.values())[0]  # one input expected
        normalization_result = normalization(input_data)

        return self.create_output_dict({"DataOutNormalization": normalization_result})



from typing import Dict

from rdflib import Graph, Namespace


class KGSchema:
    """
    Stores a Graph object and some metadata corresponding to a KG schema
    """

    def __init__(self, path: str, namespace: str, namespace_prefix: str):
        self.path = path  # path of the KG schema definition, can be local or remote
        self.namespace = Namespace(namespace)
        self.namespace_prefix = namespace_prefix

        self.kg = Graph(bind_namespaces="rdflib")
        self.kg.parse(self.path, format="n3")

    @classmethod
    def from_schema_info(cls, schema_info: Dict[str, str]):
        return cls(schema_info["path"], schema_info["namespace"], schema_info["namespace_prefix"])



from classes import ExeKG
from utils.cli_utils import input_pipeline_info

if __name__ == "__main__":
    pipeline_name, input_data_path = input_pipeline_info()
    kg_schema_name = ExeKG.input_kg_schema_name()
    exe_kg = ExeKG(kg_schema_name=kg_schema_name)
    exe_kg.start_pipeline_creation(pipeline_name, input_data_path)
    exe_kg.save_created_kg(f"kg/{pipeline_name}.ttl")



from .entity import Entity


class DataEntity(Entity):
    """
    Abstraction of owl:class DataEntity.

    ❗ Important for contributors ❗
    The fields that contain "_" are by convention the snake-case conversions of the equivalent camel-case property names in the KG.
    e.g. has_source field corresponds to hasSource property in the KG.
    This is necessary for automatically mapping KG properties to Python object fields while parsing the KG.
    """

    def __init__(
        self,
        iri: str,
        parent_entity: Entity,
        has_source_value: str = None,
        has_data_semantics_iri: str = None,
        has_data_structure_iri: str = None,
        has_reference: str = None,
    ):
        super().__init__(iri, parent_entity)
        self.has_source = has_source_value  # used as column name to retrieve data from the pipeline's input file
        self.has_data_semantics = has_data_semantics_iri  # IRI of KG entity of type DataSemantics
        self.has_data_structure = has_data_structure_iri  # IRI of KG entity of type DataStructure
        self.has_reference = has_reference  # reference to another data entity in the KG, expecting an IRI



from typing import Dict, Union

from classes.data_entity import DataEntity
from classes.entity import Entity
from classes.task import Task
from rdflib import RDF, XSD, Graph, Literal, Namespace, URIRef
from utils.query_utils import (get_data_properties_by_entity_iri,
                               get_first_query_result_if_exists)


def add_instance(kg: Graph, entity_instance: Entity) -> None:
    """
    Adds entity instance to KG only if its parent entity exists and there is no instance with the same IRI
    Args:
        kg: Graph object to add to
        entity_instance: the entity instance to create
    """
    if entity_instance.parent_entity and (entity_instance.iri, None, None) not in kg:
        kg.add((entity_instance.iri, RDF.type, entity_instance.parent_entity.iri))


def add_relation(kg: Graph, from_entity: Entity, relation_iri: str, to_entity: Entity) -> None:
    """
    Adds relation between 2 given entities to KG
    Args:
        kg: Graph object to add to
        from_entity: relation source
        relation_iri: IRI that connects the 2 given entities
        to_entity: relation destination
    """
    kg.add(
        (
            from_entity.iri,
            URIRef(relation_iri),
            to_entity.iri,
        )
    )


def add_literal(kg: Graph, from_entity: Entity, relation_iri: str, literal: Literal) -> None:
    """
    Adds relation between a given entity and a given literal to KG
    Args:
        kg: Graph object to add to
        from_entity: relation source
        relation_iri: IRI that connects the given entity with the given literal
        literal: literal to add to Graph object
    """
    kg.add((from_entity.iri, URIRef(relation_iri), literal))


def add_instance_from_parent_with_relation(
    namespace: Namespace,
    kg: Graph,
    parent_entity: Entity,
    relation_iri: str,
    related_entity: Entity,
    instance_name: str,
) -> Entity:
    """
    Creates an entity object based on the arguments and calls add_instance() and add_relation() to create a new entity instance and relation
    Args:
        namespace: namespace for the new instance
        kg: Graph object to add to
        parent_entity: parent entity for the new instance
        relation_iri: IRI that connects the given related_entity with the new instance
        related_entity: relation source
        instance_name: name for the new instance

    Returns:
        Entity: object containing the new entity instance's basic info
    """
    entity_iri = namespace + instance_name
    instance = Entity(entity_iri, parent_entity)

    add_instance(kg, instance)
    add_relation(kg, related_entity, relation_iri, instance)

    return instance


def name_instance(
    task_type_dict: Dict[str, int],
    method_type_dict: Dict[str, int],
    parent_entity: Entity,
) -> Union[None, str]:
    """
    Creates a unique name for a new instance by concatenating the parent entity's name (which is the instance type) with a number
    Also increments the relevant number of the corresponding dict
    Args:
        task_type_dict: contains pairs of task types and numbers
        method_type_dict: contains pairs of method types and numbers
        parent_entity: instance's parent entity

    Returns:
        str: name to be given to the new instance
        None: if the type of the given parent entity is not equal with "AtomicTask" or "AtomicMethod"
    """
    if parent_entity.type == "AtomicTask":
        entity_type_dict = task_type_dict
    elif parent_entity.type == "AtomicMethod":
        entity_type_dict = method_type_dict
    else:
        print("Error: Invalid parent entity type")
        return None

    instance_name = parent_entity.name + str(entity_type_dict[parent_entity.name])
    entity_type_dict[parent_entity.name] += 1
    return instance_name


def add_data_entity_instance(
    kg: Graph,
    data: Entity,
    top_level_kg: Graph,
    top_level_schema_namespace: Namespace,
    data_entity: DataEntity,
) -> None:
    """
    Adds data entity instance to kg with the necessary relations
    Args:
        kg: Graph object to add to
        data: object representing top-level DataEntity class in KG
        top_level_kg: KG corresponding to the top-level KG schema
        top_level_schema_namespace: namespace of the top-level KG schema
        data_entity: data entity to add
    """
    add_instance(kg, data_entity)

    if data_entity.has_source:
        has_source_iri, range_iri = get_first_query_result_if_exists(
            get_data_properties_by_entity_iri, data.iri, top_level_kg
        )

        source_literal = Literal(
            lexical_or_value=data_entity.has_source,
            datatype=range_iri,
        )

        add_literal(kg, data_entity, has_source_iri, source_literal)

    if data_entity.has_data_structure:
        add_relation(
            kg,
            data_entity,
            top_level_schema_namespace.hasDataStructure,
            Entity(data_entity.has_data_structure),
        )

    if data_entity.has_data_semantics:
        add_relation(
            kg,
            data_entity,
            top_level_schema_namespace.hasDataSemantics,
            Entity(data_entity.has_data_semantics),
        )

    if data_entity.has_reference:
        add_relation(
            kg,
            data_entity,
            top_level_schema_namespace.hasReference,
            Entity(data_entity.has_reference),
        )


def add_and_attach_data_entity(
    kg: Graph,
    data: Entity,
    top_level_kg: Graph,
    top_level_schema_namespace: Namespace,
    data_entity: DataEntity,
    relation: URIRef,
    task_entity: Task,
) -> None:
    """
    Adds data entity instance to kg with the necessary relations, and attaches it to the given task
    Args:
        kg: Graph object to add to
        data: object representing top-level DataEntity class in KG
        top_level_kg: KG corresponding to the top-level KG schema
        top_level_schema_namespace: namespace of the top-level KG schema
        data_entity: data entity to add
        relation: IRI of relation to add
        task_entity: task to attach the data entity to
    """
    add_data_entity_instance(kg, data, top_level_kg, top_level_schema_namespace, data_entity)
    add_relation(kg, task_entity, relation, data_entity)


def create_pipeline_task(
    top_level_schema_namespace: Namespace,
    bottom_level_schema_namespace: Namespace,
    parent_entity: Entity,
    kg: Graph,
    pipeline_name: str,
    input_data_path: str,
) -> Task:
    """
    Adds instance of pipeline task to kg
    Args:
        top_level_schema_namespace: namespace of the top-level KG schema
        bottom_level_schema_namespace: namespace of the bottom-level KG schema
        parent_entity: parent entity of pipeline instance
        kg: Graph object to add to
        pipeline_name: name for the pipeline
        input_data_path: path for the input data to be used by the pipeline's tasks

    Returns:
        Task: created pipeline task
    """
    pipeline = Task(bottom_level_schema_namespace + pipeline_name, parent_entity)
    add_instance(kg, pipeline)

    input_data_path_literal = Literal(lexical_or_value=input_data_path, datatype=XSD.string)
    add_literal(kg, pipeline, top_level_schema_namespace.hasInputDataPath, input_data_path_literal)

    return pipeline



from typing import List, Tuple

from classes.data_entity import DataEntity
from classes.entity import Entity
from rdflib import Namespace


def get_input_for_existing_data_entities(
    existing_data_entity_list: List[DataEntity],
) -> List[DataEntity]:
    """
    Asks user to choose data entities from an existing list
    Args:
        existing_data_entity_list: contains DataEntity objects for the user to choose from

    Returns:
        List[DataEntity]: contains the chosen DataEntity objects
    """
    if not existing_data_entity_list:
        return []

    chosen_data_entity_list = []
    print("Choose input for the task from existing data entities:")
    while True:
        for i, data_entity in enumerate(existing_data_entity_list):
            print(f"\t{str(i)}. {data_entity.name}")
        print(f"\t{str(-1)}. Continue")
        chosen_data_entity_i = int(input())
        if chosen_data_entity_i == -1:
            break

        chosen_data_entity_list.append(existing_data_entity_list[chosen_data_entity_i])

    return chosen_data_entity_list


def get_input_for_new_data_entities(
    data_semantics_list: List[Entity], data_structure_list: List[Entity], namespace: Namespace, data_entity: Entity
) -> List[DataEntity]:
    """
    Asks user to specify info of new data entities and creates relevant objects
    Args:
        data_semantics_list: contains data semantics for the user to choose from
        data_structure_list: contains data structures for the user to choose from
        namespace: KG schema namespace to use when initializing the new entities
        data_entity: Entity object to assign as parent entity of the new entities

    Returns:
        List[DataEntity]: contains the created DataEntity objects
    """
    data_entities = []

    prompt = "Enter input columns, then 'quit' when done: "
    source = input(prompt)
    while source != "quit":
        new_data_entity = DataEntity(namespace + source, data_entity, source)

        print(f"Choose data semantics for {source}:")
        for i, t in enumerate(data_semantics_list):
            print(f"\t{str(i)}. {t.name}")
        chosen_data_semantics_id = int(input())
        new_data_entity.has_data_semantics = data_semantics_list[chosen_data_semantics_id].iri

        print(f"Choose data structure for {source}:")
        for i, t in enumerate(data_structure_list):
            print(f"\t{str(i)}. {t.name}")
        chosen_data_structure_id = int(input())
        new_data_entity.has_data_structure = data_structure_list[chosen_data_structure_id].iri

        data_entities.append(new_data_entity)

        source = input(prompt)

    return data_entities


def input_pipeline_info() -> Tuple[str, str]:
    """
    Asks user to provide a name for the pipeline and a path for the input data
    Returns:
        Tuple[str, str]: contains the provided strings
    """
    pipeline_name = input("Enter a name for the pipeline: ")
    input_data_path = input("Enter a path for the input data: ")

    return pipeline_name, input_data_path



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exe_kg_lib',
 'exe_kg_lib.classes',
 'exe_kg_lib.classes.tasks',
 'exe_kg_lib.cli',
 'exe_kg_lib.utils',
 'exe_kg_lib.utils.task_utils']

package_data = \
{'': ['*']}

install_requires = \
['black[d]>=22.10.0,<23.0.0',
 'matplotlib>=3.6.2,<4.0.0',
 'mkdocs>=1.4.2,<2.0.0',
 'pandas>=1.5.2,<2.0.0',
 'rdflib>=6.2.0,<7.0.0',
 'scikit-learn>=1.1.3,<2.0.0',
 'typer>=0.7.0,<0.8.0']

setup_kwargs = {
    'name': 'exe-kg-lib',
    'version': '1.0.7',
    'description': 'Library for executable ML pipelines represented by KGs.',
    'long_description': "# ExeKGLib\n\n![PyPI](https://img.shields.io/pypi/v/exe-kg-lib)\n![Python](https://img.shields.io/badge/python-v3.8+-blue.svg)\n![Poetry](https://img.shields.io/badge/poetry-v1.2.2-blue)\n[![Code style: black][black-badge]][black]\n[![License](https://img.shields.io/badge/license-AGPL%203.0-blue)](https://www.gnu.org/licenses/agpl-3.0.en.html)\n\nPython library for conveniently constructing and executing Machine Learning (ML) pipelines represented by Knowledge Graphs (KGs).\n\nDetailed information (installation, documentation etc.) about the library can be found in [its website](https://boschresearch.github.io/ExeKGLib/) and basic information is shown below.\n\n## Overview\n\n[//]: # (--8<-- [start:overview])\nThe functionality of this Python library can be divided in the below two parts:\n\n1. **Executable KG construction**: An executable KG representing an ML pipeline is constructed as per user's input (programmatically or via CLI) based on the KG schemas. The construction is done by sequentially creating pairs of instances of [ds:AtomicTask](https://nsai-uio.github.io/ExeKGOntology/OnToology/ds_exeKGOntology.ttl/documentation/index-en.html#AtomicTask) and [ds:AtomicMethod](https://nsai-uio.github.io/ExeKGOntology/OnToology/ds_exeKGOntology.ttl/documentation/index-en.html#AtomicMethod) sub-classes, and their properties. The definition of these sub-classes can be found in the [bottom-level KG schemas](#bottom-level-kg-schemas). After each KG component is built, it is validated using the KG schemas and added to an RDFLib `Graph` object. The KG is finally saved in Turtle format.\n2. **ML pipeline execution**: The executable KG is parsed using RDFLib and queried using SPARQL to retrieve its ML pipeline. The pipeline's ordered tasks are sequentially mapped to Python objects that include an implemented `run_method()` Python method which is then invoked. This is as an abstract method of the _Task_ class that is implemented by its bottom-level children classes.\n\nThe different implementations of `run_method()` correspond to the [ds:AtomicMethod](https://nsai-uio.github.io/ExeKGOntology/OnToology/ds_exeKGOntology.ttl/documentation/index-en.html#AtomicMethod) bottom-level sub-classes that are defined in the Visualization, Statistics, and ML KG schemas. The method categories are described below.\n\n1. **Visualization**: This is a set of methods for visualization, including two types: (1) The plot canvas methods that define the plot size and layout. (2) The various kinds of plot methods (line plot, scatter plot, bar plot, etc.). These methods use matplotlib to visualize data.\n2. **Statistics and Feature Engineering**: This includes methods for statistical analysis and feature engineering like IQR calculation, mean and std-deviation calculation, etc., which can then form complex methods like outlier detection method and normalization method.\n3. **Machine Learning**: This is a group of methods that support ML algorithms like Linear Regression, MLP, and k-NN and helper functions that perform e.g. data splitting and ML model performance calculation.\n\nThis library is part of the following paper submitted to ESWC 2023:<br>\n_Klironomos A., Zhou B., Tan Z., Zheng Z., Gad-Elrab M., Paulheim H., Kharlamov E.: **ExeKGLib: A Python Library for Machine Learning Analytics based on Knowledge Graphs**_\n\n[//]: # (--8<-- [end:overview])\n\n## Getting started\n\n[//]: # (--8<-- [start:gettingstarted])\nThe library is available as a [PyPi package](https://pypi.org/project/exe-kg-lib/).\n\nTo download, run `pip install exe-kg-lib`.\n\n[//]: # (--8<-- [end:gettingstarted])\n\n## Usage\n\n[//]: # (--8<-- [start:usage])\n### Creating an executable KG\n#### Via CLI\n1. Run `python kg_construction.py`.\n2. Follow the input prompts.\n\n#### Via code\nSee the [provided examples](examples/).\n\n### Executing a generated KG\nRun `python kg_execution.py [kg_file_path]`.\n\n[//]: # (--8<-- [end:usage])\n\n## Installation\nSee the [installation page](https://boschresearch.github.io/ExeKGLib/installation/) of the library's website.\n\n## Adding a new ML-related task and method\n\n[//]: # (--8<-- [start:extending])\nTo perform this type of library extension, there are 3 required steps:\n\n1. Selection of a relevant bottom-level KG schema (Statistics, ML, or Visualization) according to the type of the new task and method.\n2. Addition of new semantic components (entities, properties, etc) to the selected KG schema.\n3. Addition of a Python class to the corresponding module of `exe_kg_lib.classes.tasks` package.\n\nFor steps 2 and 3, refer to the [relevant page](https://boschresearch.github.io/ExeKGLib/adding-new-task-and-method/) of the library's website.\n\n[//]: # (--8<-- [end:extending])\n\n## Documentation\nSee the _Code Reference_ and _Development_ sections of the [library's website](https://boschresearch.github.io/ExeKGLib/).\n\n## External resources\n\n[//]: # (--8<-- [start:externalresources])\n### Top-level KG schemas\n- [Data Science KG schema](https://w3id.org/def/exekg-ds)\n\n### Bottom-level KG schemas\n- [Visualization KG schema](https://w3id.org/def/exekg-visu)\n- [Statistics KG schema](https://w3id.org/def/exekg-stats)\n- [Machine Learning KG schema](https://w3id.org/def/exekg-ml)\n\nThe above KG schemas are included in the [ExeKGOntology repository](https://github.com/nsai-uio/ExeKGOntology).\n\n### Dataset used in code examples\nThis dataset (located in `exe_kg_lib/examples/data/dummy_data.csv`) was generated using the `sklearn.datasets.make_classification()` function of the [scikit-learn Python library](https://scikit-learn.org/).\n\n[//]: # (--8<-- [end:externalresources])\n\n## License\n\nExeKGLib is open-sourced under the AGPL-3.0 license. See the\n[LICENSE](LICENSE.md) file for details.\n\n<!-- URLs -->\n[black-badge]: https://img.shields.io/badge/code%20style-black-000000.svg\n[black]: https://github.com/psf/black\n[ci-badge]: https://github.com/boschresearch/ExeKGLib/actions/workflows/ci.yaml/badge.svg\n[ci]: https://github.com/boschresearch/ExeKGLib/actions/workflows/ci.yaml\n[docs-badge]: https://img.shields.io/badge/docs-gh--pages-inactive\n[docs]: https://github.com/boschresearch/ExeKGLib/tree/gh-pages\n[license-badge]: https://img.shields.io/badge/License-All%20rights%20reserved-informational\n[license-url]: https://pages.github.boschdevcloud.com/bcai-internal//latest/license\n[pre-commit-badge]: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n[pre-commit]: https://github.com/pre-commit/pre-commit\n",
    'author': 'Antonis Klironomos',
    'author_email': 'antonis.klironomos@de.bosch.com',
    'maintainer': 'Antonis Klironomos',
    'maintainer_email': 'antonis.klironomos@de.bosch.com',
    'url': 'https://boschresearch.github.io/ExeKGLib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import sys

from classes import ExeKG

if __name__ == "__main__":
    try:
        exe_kg_path = sys.argv[1]
    except IndexError:
        exe_kg_path = "../examples/pipelines/MLPipeline.ttl"

    exe_kg = ExeKG(input_exe_kg_path=exe_kg_path)
    exe_kg.execute_pipeline()



from classes.entity import Entity
from classes.task import Task
from utils.task_utils.ml_utils import *

"""
❗ Important for contributors: See the package's README.md before extending the code's functionality.
"""


class ConcatenationConcatenationMethod(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        concatenation_result = concatenation(list(input_dict.values()))
        return self.create_output_dict({"DataOutConcatenatedData": concatenation_result})


class DataSplittingDataSplittingMethod(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)
        self.has_split_ratio = None

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_x = input_dict["DataInDataSplittingX"]
        input_y = input_dict["DataInDataSplittingY"]
        train_x, test_x, train_y, test_y = data_splitting(input_x, input_y, self.has_split_ratio)

        return self.create_output_dict(
            {
                "DataOutSplittedTrainDataX": train_x,
                "DataOutSplittedTestDataX": test_x,
                "DataOutSplittedTrainDataY": train_y,
                "DataOutSplittedTestDataY": test_y,
            }
        )


class TrainKNNTrain(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_x = input_dict["DataInTrainX"]
        input_y = input_dict["DataInTrainY"]
        model, predicted_y = k_nn_train(input_x, input_y)

        return self.create_output_dict({"DataOutTrainModel": model, "DataOutPredictedValueTrain": predicted_y})


class TestKNNTest(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        model = input_dict["DataInTestModel"]
        input_x = input_dict["DataInTestX"]

        predicted_y = k_nn_test(model, input_x)
        return self.create_output_dict({"DataOutPredictedValueTest": predicted_y})


class TrainLRTrain(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_x = input_dict["DataInTrainX"]
        input_y = input_dict["DataInTrainY"]

        model, predicted_y = lr_training(input_x, input_y)

        return self.create_output_dict({"DataOutModel": model, "DataOutPredictedValueTrain": predicted_y})


class TestLRTest(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        model = input_dict["DataInModel"]
        input_x = input_dict["DataInTestX"]

        predicted_y = lr_testing(model, input_x)
        return self.create_output_dict({"DataOutPredictedValueTest": predicted_y})


class TrainMLPTrain(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        input_x = input_dict["DataInTrainX"]
        input_y = input_dict["DataInTrainY"]

        model, predicted_y = mlp_train(input_x, input_y)

        return self.create_output_dict({"DataOutModel": model, "DataOutPredictedValueTrain": predicted_y})


class TestMLPTest(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        model = input_dict["DataInModel"]
        input_x = input_dict["DataInTestX"]

        predicted_y = mlp_test(model, input_x)
        return self.create_output_dict({"DataOutPredictedValueTest": predicted_y})


class PerformanceCalculationPerformanceCalculationMethod(Task):
    def __init__(self, iri: str, parent_entity: Entity):
        super().__init__(iri, parent_entity)

    def run_method(self, other_task_output_dict: dict, input_data: pd.DataFrame) -> dict:
        input_dict = self.get_inputs(other_task_output_dict, input_data)
        real_train_y = input_dict["DataInTrainRealY"]
        real_test_y = input_dict["DataInTestRealY"]
        predicted_train_y = input_dict["DataInTrainPredictedY"]
        predicted_test_y = input_dict["DataInTestPredictedY"]

        train_error, test_error = ml_performance_calculation(
            real_train_y, real_test_y, predicted_train_y, predicted_test_y
        )

        return self.create_output_dict({"DataOutMLTrainErr": train_error, "DataOutMLTestErr": test_error})
